/**
 * Form Component
 */
import BookingRequest from '../types/booking-request.mjs';
import HttpRequest from './HttpRequest.mjs';
import Service from "./service.mjs";
import Loader from "./loader.mjs";
import AddressAutoCompleteResult from "./AddressAutoCompleteResult.mjs";
import GoogleAutocomplete from "../services/google-autocomplete.mjs";
import PhoneNumberValidator from "../services/phone-number-validator.mjs";
import {FieldValidator} from "../services/field-validator.mjs";
import ErrorMessage from "../services/error-message.mjs";
import log from "../services/log.mjs";
import errorMessage from "../services/error-message.mjs";
import GoogleAnalytics from "../services/google-analytics.mjs";
import {clone} from "chart.js/helpers";
import BookingFormStepGenerator from "./booking-form-step-generator.mjs";
import ServiceGroup from "./service-group.mjs";

/**
 * Booking Form
 * @type {Form}
 */
class BookingForm {

    /**
     * Form element
     * @type {{}}
     */
    formElement = {};

    dirty = true;

    formFieldValidationStatus = {};

    /**
     * @type CustomEvent
     */
    formFieldValidationEvent = null;

    /**
     * @type CustomEvent
     */
    servicesLoadedEvent = null;

    usedJimsBefore = false;

    useGoogleAutocomplete = false;

    googleAnalytics = null;

    GoogleAutocomplete = null;

    options = {};

    serviceList = [];

    groups = [];

    addressSuggestions = [];

    endpoints = {
        services: 'jims-wpcb/v1/service/division/services',
        countries: 'jims-wpcb/v1/service/countries',
        address: 'jims-wpcb/v1/customer-booking/address',
        details: 'jims-wpcb/v1/customer-booking/address/details',
        fsePhoto: 'jims-wpcb/v1/customer-booking/url/fse-photo',
        create: {
            'direct': 'jims-wpcb/v1/customer-booking/create',
            'step1': 'jims-wpcb/v1/customer-booking/create?booking_step=1',
            'step2': 'jims-wpcb/v1/customer-booking/create?booking_step=2',
        },
    }

    messages = {
        'services_loading': 'Loading Services...',
        'search': 'Searching for Franchisee in your Area...',
        'submitting': 'Submitting your request...',
    };

    errorMessages = {
        text: 'Please enter a valid text',
        email: 'Please enter a valid email',
        phone: 'Please enter a valid phone number'
    };

    /**
     * Progress
     * @type {number}
     */
    progress = 0;

    /**
     *
     * @type {null}
     */
    bookingRequest = null;

    /**
     * Required fields
     */
    requiredFields = [
        'firstname',
        'lastname',
        'email',
        'phone',
        'address',
    ];

    prefixClass = 'jims-wpcb';


    /**
     *
     * @type {{Object:BookingFormDivisionStep|BookingFormServiceStep|BookingFormInfoStep}}
     */
    steps = {};

    /**
     * Current step
     * @type {number}
     */
    currentStep = 0;


    navButtons = {};


    countries = [];

    deviceState = {};

    /**
     *
     * @param formElement, form element
     * @param options object
     * @param useGoogleAutocomplete boolean
     */
    constructor(formElement, options, useGoogleAutocomplete, overrides = {}) {
        this.formElement = formElement;
        this.options = this.applyOverrides(options, overrides);
        this.useGoogleAutocomplete = useGoogleAutocomplete;

        //if window is not defined, define it
        if (!window) {
            const window = {};
        }

        //assume that the Google Analytics script is already loaded
        this.googleAnalytics = new GoogleAnalytics(window);

        this.initialize();
    }


    initialize() {
        //set up loader
        this.loader = new Loader(`.${this.getSelector()} .${this.prefixClass}-loader-wrapper`);

        // this.loadSteps();
        if (this.options.enable_popup === '1') {

            if (document.querySelector('html').hasAttribute('class')
                && document.querySelector('html').getAttribute('class').indexOf('avia-') > -1) {
                //move formElement to the end of the body
                document.body.appendChild(this.formElement);
            } else {
                //append the formElement to the start of the body
                document.body.insertBefore(this.formElement, document.body.firstChild);
            }

            this.listenForPopup();
        }

        //load division services
        this.loadServices();

        this.navButtons = {
            prev: this.formElement.querySelector(`.${this.prefixClass}-button--prev`),
            next: this.formElement.querySelector(`.${this.prefixClass}-button--next`),
            submit: this.formElement.querySelector(`.${this.prefixClass}-button--submit`)
        }

        //set up default loader message
        this.formElement.querySelector('.loader-message').innerHTML = this.messages.services_loading;


        this.formFieldValidationEvent = new CustomEvent('jims-wpcb-form-field-validation', {
            bubbles: true,
            detail: {
                field: null,
                status: false,
            }
        });

        this.servicesLoadedEvent = new CustomEvent('jims-wpcb-services-loaded', {
            bubbles: true,
        });

        // Important Note: the order of these functions is important
        // for example setupProgress must be called first
        // then setupAddressAutocomplete
        // then setupCountries
        // this is because setupCountries depends on Autocomplete
        this.setupProgress();
        this.setupAddressAutocomplete();
        this.setupCountries();
        this.setupPhone();


        //add event listeners
        this.setupEvents();
        this.setupSearch();
        this.setupValidation();
        this.calculateDeviceHeight();
    }

    removeURLHash() {
        window.location.hash = '';
    }

    listenForPopup() {
        let hash = window.location.hash;

        if (hash === '#jims') {
            //open the popup
            this.formElement.querySelector(`.${this.prefixClass}-popup-container`).classList.add('open');
        }

        //listen for hash change
        window.addEventListener('hashchange', (evt) => {
            if (window.location.hash === '#jims') {
                this.formElement.querySelector(`.${this.prefixClass}-popup-container`).classList.add('open');

                //remove hash
                history.pushState("", document.title, window.location.pathname + window.location.search);
            }
        });

        //listen for close button
        this.formElement.querySelector(`.${this.prefixClass}-popup-close`).addEventListener('click', (evt) => {
            this.formElement.querySelector(`.${this.prefixClass}-popup-container`).classList.remove('open');

            //remove hash
            history.pushState("", document.title, window.location.pathname + window.location.search);
        });
    }

    getSelector() {
        let classes = [];
        for (let cssClass of this.formElement.classList.values()) {
            classes.push(cssClass);
        }

        return classes.join('.');
    }

    applyOverrides(options, overrides) {
        //clone option is important, in other case it overrides the wpcb variables
        const overrideOptions = clone(options);

        if (overrides.division_id && (overrideOptions.bookingFormDivisionId !== overrides.division_id)) {
            overrideOptions.bookingFormDivisionId = overrides.division_id;
        }

        if (overrides.service_ids) {
            overrideOptions.bookingFormServiceIds = overrides.service_ids;
        }

        log.info(overrideOptions, 'overrideOptions');

        //delete properties from overrides
        delete overrides.division_id;
        delete overrides.service_ids;

        //apply rest of the overrides
        for (const [key, value] of Object.entries(overrides)) {
            overrideOptions[key] = value;
        }

        return overrideOptions;
    }

    loadSteps() {
        const generator = new BookingFormStepGenerator();

        //find all divs with class jims-wpcb-form-step
        let steps = this.formElement.querySelectorAll(`.${this.prefixClass}__step`);
        steps.forEach((step, index) => {
            if (step.dataset.name && step.dataset.name.length > 0) {
                this.steps[step.dataset.name] = generator.createStep(step.dataset.name, step, this.formElement);
            }
        });
    }

    calculateDeviceHeight() {
        let wrapper = this.formElement.querySelector('.jims-wpcb-form__wrapper');

        if (this.deviceState.formHeight === undefined) {
            this.deviceState.formHeight = wrapper.getBoundingClientRect().height;
        }

        let rect = wrapper.getBoundingClientRect();

        //detect orientation
        let orientation = window.matchMedia("(orientation: portrait)");

        if (orientation.matches && window.innerWidth < 768) {
            wrapper.style.maxHeight = `${window.innerHeight - rect.y}px`;
        } else {
            wrapper.style.maxHeight = `inherit`;
        }
    }

    setupCountries() {
        let http = new HttpRequest(this.options.rest_url);
        let countriesSelect = this.formElement.querySelector('select[name="country"]');
        const config = this.options;

        // if (countriesSelect === null) { return false; }

        http.get(this.endpoints.countries).then((response) => {
            this.countries = response;

            this.countries.forEach((country) => {
                let option = document.createElement('option');
                option.value = country.symbol;
                option.innerText = country.name;

                if (config.country && country.symbol === config.country) {
                    option.setAttribute('selected', 'selected');
                    this.setFlag(country.symbol);
                }

                countriesSelect.appendChild(option);
            });
        });

        countriesSelect.addEventListener('change', (evt) => {
            this.setFlag(evt.target.value);
            if (this.useGoogleAutocomplete) {
                this.GoogleAutocomplete.setCountry(evt.target.value);
            }
        });
    }

    setFlag(countryCode) {
        let countryImage = countryCode.toLowerCase();
        this.formElement.querySelector('.icon-country img').setAttribute('src', `${this.options.wpcb_url}assets/dist/img/${countryImage}.png`);
    }

    /**
     * Refreshes the address suggestions
     */
    setupAddressAutocomplete() {
        const addressInput = this.formElement.querySelector('input[name="address"]');

        if (this.useGoogleAutocomplete) {
            this.GoogleAutocomplete = new GoogleAutocomplete(
                addressInput,
                this.formElement,
                {
                    countryCode: this.options.country,
                },
                this.options.google_api_key
            );
        } else {
            let handler = null;
            addressInput.addEventListener('keyup', (evt) => {
                handler = setTimeout(() => {
                    clearTimeout(handler);
                    this.getAddressSuggestions(evt.target.value);
                }, 500);
            });
        }
    }

    /**
     * Sets up the phone number
     */
    setupPhone() {
        const phoneElement = this.formElement.querySelector('.title-container .phone');
        const notificationPhoneElement = this.formElement.querySelector('.call-us');
        const data = ['href', `tel:${this.options.phone}`.trim().replace(/\s/g, '')];

        phoneElement.parentElement.setAttribute(...data);
        notificationPhoneElement.setAttribute(...data);

        notificationPhoneElement.querySelector('span.phone').innerHTML = this.options.phone;
        phoneElement.innerHTML = this.options.phone;
    }

    /**
     * Requests address autocomplete suggestions from findajim service
     * @param address
     * @param addressInput
     */
    getAddressSuggestions(address) {
        const httpRequest = new HttpRequest(this.options.rest_url);
        httpRequest.post(this.endpoints.address, {
            "address": address,
            "country_code": this.formElement.querySelector('select[name="country"]').value
        }).then((response) => {
            this.addressSuggestions = response;
            this.setAddressSuggestions();
        });
    }

    /**
     * Adds options to auto-complete for address
     */
    setAddressSuggestions() {
        const autocompleteElementContainer = this.formElement.querySelector(`.${this.prefixClass}__autocomplete`);
        autocompleteElementContainer.innerHTML = '';
        autocompleteElementContainer.classList.add('open');

        //populate datalist with suggestions
        this.addressSuggestions.forEach((address) => {
            let autocompleteElement = new AddressAutoCompleteResult(
                address,
                this.prefixClass,
                this.formElement.querySelector('input[name="address"]'),
                autocompleteElementContainer
            );
            autocompleteElementContainer.append(autocompleteElement.render());
        });
    }


    /**
     * Initialize search capabilities
     */
    setupSearch() {
        this.seachbar = this.formElement.querySelector('.search-bar');
        let handler = null;
        this.seachbar.addEventListener('keyup', (evt) => {
            //set timeout and cancel previous timeout
            handler = setTimeout(() => {
                clearTimeout(handler);
                this.searchServices(evt.target.value);
            }, 500);
        });
    }

    searchServices(keyword) {
        //hide all services
        this.serviceList.forEach((service) => {
            service.hide();

            if (service.name.toLowerCase().includes(keyword.toLowerCase())) {
                service.show();
            }
        });
    }

    setupValidation() {
        this.navButtons.submit.classList.add('disabled');
        let form = this.formElement.querySelector('form');
        let PhoneValidator = new PhoneNumberValidator(this.options.rest_url);
        let countryCode = this.formElement.querySelector('select[name="country"]').value;

        if (countryCode.length === 0) {
            countryCode = this.options.country;
        }


        //set a form onchange observer
        form.addEventListener('change', (e) => {
            let field = e.target;

            //check if field nodeName is INPUT | SELECT | TEXTAREA
            if (!['INPUT', 'TEXTAREA'].includes(field.nodeName)) {
                return;
            }

            //if target is required
            if (field.hasAttribute('required')) {
                field.classList.add('touched');
                //if field is empty
                if (field.type === 'tel') {
                    PhoneValidator.validate(
                        countryCode,
                        field.value,
                    ).then((response) => {
                        this.formFieldValidationStatus[field.name] = response.result;
                        this.formElement.dispatchEvent(this.formFieldValidationEvent);
                    });
                } else {
                    let status = FieldValidator.validate(field.type, field.value);
                    this.formFieldValidationStatus[field.name] = status;
                    this.formElement.dispatchEvent(this.formFieldValidationEvent);
                }
            }
        });
    }

    reset() {
        this.formElement.querySelector('form').reset();
        this.formElement.querySelector('.jims-wpcb-loader-wrapper').classList.remove('open', 'success', 'error');
        this.navButtons.submit.classList.add('disabled');
        this.resetNotification();
        this.usedJimsBefore = false;

        this.serviceList.forEach((service) => {
            service.update(service.id, service.name, false);
        });

        //remove touched class from all fields
        this.formElement.querySelectorAll('.touched').forEach((field) => {
            field.classList.remove('touched');
        });

        this.prevStep();
    }

    /**
     *
     * @param status
     * @param requiredFields
     * @returns {boolean}
     */
    isDirty() {
        //filter formFieldValidationStatus for false values
        let dirtyFields = Object.values(this.formFieldValidationStatus).filter((value) => {
            return value === false;
        });

        return (dirtyFields.length > 0) || !(this.formElement.querySelectorAll('.touched').length === this.requiredFields.length);
    }

    /**
     *
     */
    setupEvents() {
        this.navButtons.submit.addEventListener('click', (evt) => {
            evt.preventDefault();
            this.formElement.querySelector('form').dispatchEvent(new Event('submit', {
                cancelable: true,
                bubbles: true
            }));

            return false;
        });

        this.formElement.querySelector('.try-again').addEventListener('click', (evt) => {
            evt.preventDefault();
            this.resetNotification();
            return false;
        });

        //Resets the form and clears all fields
        this.formElement.querySelector('.booking-action.new').addEventListener('click', (evt) => {
            this.reset();
        });

        this.formElement.querySelector('.used-before').addEventListener('click', (evt) => {
            if (evt.target.classList.contains('yes')) {
                this.usedJimsBefore = true;
            }

            // this.loader.classList.remove('used');
            this.loader.modifyClass(['used'], 'remove');
            this.setFormNotification(this.messages.submitting);

            //and run processFormStep1
            this.processFormStep1(this.previous_response, this._fields, this.usedJimsBefore).then((response) => {
                this.processFormDirect(response);
            }).catch((error) => {
                log.info(error, 'processformstep1');
                this.setErrorMessageNotification(error);
            });
        });

        //Submit form
        this.formElement.querySelector('form').addEventListener('submit', (evt) => {
            evt.preventDefault();
            this.googleAnalytics.sendEvent('form', 'submit', 'booking-submit', 1)

            this.validateAndSubmitForm();
            return false;
        });

        //Live validation event
        this.formElement.addEventListener('jims-wpcb-form-field-validation', (evt) => {
            this.dirty = this.isDirty();

            if (!this.dirty) {
                this.navButtons.submit.classList.remove('disabled');
            } else {
                this.navButtons.submit.classList.add('disabled');
            }

            Object.keys(this.formFieldValidationStatus).forEach((name) => {
                let field = this.formElement.querySelector(`[name="${name}"]`);
                if (this.formFieldValidationStatus[name]) {
                    field.classList.remove('error');
                } else {
                    field.classList.add('error');
                }
            });
        });

        this.formElement.addEventListener('jims-wpcb-services-loaded', () => {
            this.loader.removeServiceLoader();
        });

        //Update selected services
        this.formElement.addEventListener('change', (evt) => {
            this.loadSelectedServices();
        });

        this.navButtons.prev.addEventListener('click', (evt) => {
            this.prevStep();
        });

        this.navButtons.next.addEventListener('click', (evt) => {
            this.nextStep();
        });

        this.formElement.querySelector('.try-again').addEventListener('mouseup', (evt) => {
            this.loader.removeError();
            this.usedJimsBefore = false;
        });


        //Tooltip
        this.formElement.querySelectorAll('.certification').forEach((cert_item) => {
            //cert item on mouseup open and focus tooltip
            cert_item.addEventListener('mouseup', (evt) => {
                let tooltip = cert_item.querySelector('.tooltip');
                tooltip.classList.add('open');
                tooltip.focus();
            });

            //cert item on focusout close tooltip
            cert_item.querySelector('.tooltip').addEventListener('focusout', (evt) => {
                cert_item.querySelector('.tooltip').classList.remove('open');
            });
        });

        if (window !== undefined) {
            window.addEventListener('resize', (evt) => {
                //recalculate form wrapper height
                this.calculateDeviceHeight();


                let rect = this.formElement.getBoundingClientRect();
                //if form is visible on the screen scroll to the top of the form
                if (rect.top < 0 && rect.bottom >= window.scrollY) {
                    //scroll to the top of the form
                    this.formElement.scrollIntoView({behavior: "instant", block: "start", inline: "start"});
                }
            });
        }
    }

    validateAndSubmitForm() {
        if (!this.dirty) {
            this.processForm();
        }
    }

    /**
     * Runs on form submission
     * @returns void
     */
    processForm() {
        this.formElement.querySelector('.jims-wpcb-loader-wrapper').classList.add('open');
        const form = this.formElement.querySelector('form');
        //firstname, lastname, email, phone, address, comments, businessName, services
        //get all required fields from form
        let addressString = form.querySelector('input[name="address"]').value;
        let selectedServices = this.getSelectedServices();

        this.getAddressObject(addressString).then((addressObject) => {

            const fields = this.getRequestFields(form, addressObject, selectedServices);
            this.BookingRequest = new BookingRequest(fields.firstname, fields.lastname, fields.email, fields.phone, fields.address, fields.comments, fields.services);
            log.info(fields, BookingRequest);


            //Send request to create booking: allocation step
            this.send(this.endpoints.create.step1, fields).then((response) => {
                log.info(response);

                //if the server returns a step
                if (response.status === 'step') {
                    this.previous_response = response;

                    //check if response.fmsData has property 'previousFSEId', which means the user probably has used Jims before
                    if (response.fms_data.payload.hasOwnProperty('previousFSEId') && response.fms_data.payload.previousFSEId > 0) {
                        this._fields = fields;

                        //prompt question
                        //on answer update form message to 'Submitting your request...'
                        this.openPromptUsedJimsBefore();
                    } else {

                        //if the response does not have property 'previousFSEId' and the server respond with multiple available FSEs, proceed to step 2
                        this.processFormStep2(response, fields, false).then((response_alloc) => {
                            this.processFormDirect(response_alloc);
                        }).catch((error) => {

                            log.info(error, 'processformstep2');
                            this.setErrorMessageNotification(error);
                        });
                    }
                } else {
                    //if there is a direct booking from the allocation request
                    this.processFormDirect(response);
                }

            }).catch((error) => {
                log.info(error, 'processform');
                this.setErrorMessageNotification(error);
            });
        });
    }

    processFormStep2(response, fields, used_jims_before) {

        let data = response;
        if (data.hasOwnProperty('fms_data')) {
            data = Array.isArray(response.fms_data) ? response.fms_data[response.fms_data.length - 1] : response.fms_data;
        }
        log.info(data);
        log.info('process 1, step 2 ', used_jims_before);

        if (data.payload.hasOwnProperty('avaliableFSEs') && data.payload.avaliableFSEs.length > 0) {
            let selectedFSE = response.fms_data.payload.avaliableFSEs[0];
            return this.send(this.endpoints.create.step2, this.getFinalRequestFields(fields, response, selectedFSE, used_jims_before));
        }

        return new Promise((resolve, reject) => {
            reject({
                statusText: 'We don\'t have anyone available to service your request at this time. If anyone becomes available in the future, they will contact you.',
            });
        });
    }

    /**
     *
     * @param response
     * @param fields
     * @param used_jims_before
     * @return {Promise<*>}
     */
    processFormStep1(response, fields, used_jims_before) {
        if (!response.fms_data.hasOwnProperty('payload')) return;

        let selectedFSE = null;
        let payload = response.fms_data.payload;
        let availableFSEs = payload.avaliableFSEs;

        log.info(payload.hasOwnProperty('previousFSEId'), payload.previousFSEId > 0, used_jims_before === false, availableFSEs.length > 1);
        if (payload.hasOwnProperty('previousFSEId') && payload.previousFSEId > 0 && used_jims_before === false && availableFSEs.length > 1) {
            availableFSEs = availableFSEs.filter((fse) => {
                return fse.only_available_previous_client === false;
            });

            log.info(availableFSEs);
        }

        log.info(payload);

        //check if response.fmsData has property 'previousFSEId'
        if (used_jims_before && payload.hasOwnProperty('previousFSEId')) {
            //if yes, check if the previousFSEId exists in the response.fmsData.availableFSEs array of objects as object.business_id
            let previousFSEId = payload.previousFSEId;

            availableFSEs.forEach((fse) => {
                if (fse.business_id === previousFSEId) {
                    selectedFSE = fse;
                }
            });
        }

        if (selectedFSE === null) {
            selectedFSE = availableFSEs[0];
        }

        log.info(selectedFSE);

        //$booking_request_id, $provider, $provider_id, confirmCodes
        //run the second step
        return this.send(this.endpoints.create.step2, this.getFinalRequestFields(fields, response, selectedFSE, used_jims_before));
    }

    /**
     * Classic form submission, without any steps and prompts
     * @param response
     */
    processFormDirect(response) {
        log.info(response);
        if (response.status === 'success') {
            this.formElement.querySelector('.jims-wpcb-loader-wrapper').classList.add('success');
            this.setFseInfo(response);
            this.googleAnalytics.sendEvent('form', 'booking', 'booking-success', 1)

            if (this.options.use_redirect === '1') {
                window.location.href = this.options.redirect_success_url;
            } else {
                this.setFormNotification(response.notification_message, 'success');
            }
        } else {
            this.formElement.querySelector('.jims-wpcb-loader-wrapper').classList.add('error');
            this.setErrorMessageNotificationForField('address', response);
        }
    }

    /**
     * Updates the error message notification
     * if there is a custom message on js frontend
     * @param fieldType string
     */
    setErrorMessageNotificationForField(fieldType, response) {
        let message = response.notification_message;

        if (response.fms_data.hasOwnProperty('errors') && response.fms_data.errors.hasOwnProperty(fieldType)) {
            message = ErrorMessage.hasOwnProperty(fieldType) ? ErrorMessage[fieldType] : message;
        }

        this.setFormNotification(message, 'error');
    }

    setErrorMessageNotification(error) {
        this.formElement.querySelector('.jims-wpcb-loader-wrapper').classList.add('error');
        this.setFormNotification(error.statusText, 'error');
    }

    setFormNotification(message) {
        let notificationElement = this.formElement.querySelector('.loader-message');
        notificationElement.innerHTML = message;
    }

    openPromptUsedJimsBefore() {
        // this.loader.classList.add('open', 'used');
        this.loader.modifyClass(['open', 'used'], 'add');
    }

    resetNotification() {
        let notificationElement = this.formElement.querySelector('.loader-message');
        notificationElement.innerHTML = this.messages.search;
    }

    /**
     * Gets the full place object from google
     * @param address
     * @returns {Promise<unknown>}
     */
    getAddressObject(address) {
        let http = new HttpRequest(this.options.rest_url);
        return http.post(this.endpoints.details, {'address': address}).then((response) => {

            let result = response;
            if (response !== null) {

                /* check if has full_address property
                and if not check if its an array
                and if yes, check if the first item has full_address property
                and reassign it to response */

                if (response.hasOwnProperty('full_address')) {
                    result = response;
                } else if (Array.isArray(response) && response.length > 0 && response[0].hasOwnProperty('full_address')) {
                    result = response[0];
                }
            }

            //replace country name with country code
            if (result !== null) {
                this.countries.forEach((country) => {
                    if (result['country'] === country.name) {
                        result['country'] = country.symbol;
                    }
                });
            }

            if (result === null) {
                //empty address field
                this.formElement.querySelector('input[name="address"]').value = '';
            }

            return result;
        });
    }

    /**
     * Returns a list of selected service ids
     * @returns Integer[]
     */
    getSelectedServices() {
        return this.serviceList.filter((service) => {
            return service.selected;
        }).map((service) => {
            return service.id;
        });
    }

    isNextButtonDisabled(selectedServices, isLastStep = null) {
        if (isLastStep === null) {
            isLastStep = this.isLastStep();
        }

        return isLastStep || selectedServices.length === 0;
    }

    isLastStep() {
        return this.currentStep === this.formElement.querySelectorAll(`.${this.prefixClass}__step`).length - 1;
    }

    setNextButtonState(selectedServices) {
        let isLastStep = this.isLastStep();

        if (this.isNextButtonDisabled(selectedServices, isLastStep)) {
            isLastStep ? this.navButtons.next.disabled = true : this.navButtons.next.classList.add('disabled');
        } else {
            this.navButtons.next.disabled = false;
            this.navButtons.next.classList.remove('disabled');
        }
    }

    setFsePhoto(fseData) {
        let request = new HttpRequest(this.options.rest_url);

        request.get(`${this.endpoints.fsePhoto}?fse_business_id=${fseData.business_id}`).then((response) => {
            this.formElement.querySelector('.fse-photo img').setAttribute('src', response.url);
        });
    }

    setSuccessVideo(isJimsFse) {
        let videoURL = 'https://www.youtube.com/embed/jclpE6EFdBk';

        if (isJimsFse) {
            videoURL = 'https://www.youtube.com/embed/cUNrjZx0n3M';
        }

        this.formElement.querySelector('.success-video iframe').setAttribute('src', videoURL);
    }

    setFseInfo(response) {
        let fseData = null;
        //check if fms_data is array
        if (Array.isArray(response.fms_data)) {
            fseData = response.fms_data[fms_data.length - 1].payload.bookedProvider;
        } else {
            fseData = response.fms_data.payload.bookedProvider;
        }

        this.setFsePhoto(fseData);
        this.setSuccessVideo(fseData.is_jims_franchisee);

        this.formElement.querySelector('.fse-name').innerText = fseData.display_name;

        if (fseData.is_jims_franchisee === false) {
            if (!!fseData.pl_insurance) {
                this.formElement.querySelector('.public-liability-amount').innerText = this.formatCurrency(fseData.pl_insurance, true);
            } else {
                this.formElement.querySelector('.public-liability').classList.add('hidden');
            }
        }

        if (!fseData.is_police_checked || fseData.is_jims_franchisee === false) {
            this.formElement.querySelector('.police-checked').classList.add('hidden');
        }

        if (!fseData.is_gold_contractor) {
            this.formElement.querySelector('.gold-contractor').classList.add('hidden');
        }

        this.setFSECallBackMessage(fseData);
        this.setFSEWorkGuarantee(fseData);
        this.setFSERatingStars(fseData);
    }

    /**
     *
     * @param amount
     * @param removeDecimal
     * @return {*|string}
     */
    formatCurrency(amount, removeDecimal = false) {
        if (isNaN(amount)) return amount;

        let AUDollar = new Intl.NumberFormat('en-US', {
            style: 'currency',
            currency: 'AUD',
        });

        if (removeDecimal) {
            //search for the dot, and remove all characters after it
            amount = parseInt(`${amount}`.split('.')[0]);
        }

        return AUDollar.format(amount).replace(/[a-zA-z]{1,2}\$/, '$');
    }

    setFSECallBackMessage(fseData) {
        //set callback message
        let callbackMessage = fseData.call_back_message;

        // find the word will and remove all characters before it
        let willIndex = callbackMessage.indexOf('will');
        callbackMessage = callbackMessage.substring(willIndex);

        //make the first letter uppercase
        callbackMessage = callbackMessage.charAt(0).toUpperCase() + callbackMessage.slice(1);

        this.formElement.querySelector('.eta_phone').innerText = callbackMessage;
    }

    setFSEWorkGuarantee(fseData) {
        let isJimsFse = fseData.is_jims_franchisee;
        let tooltip = this.formElement.querySelector('.work-guarantee-tooltip');
        let contractorTooltipHTML = `A Bizza work guarantee to the value of $200 covers this contractor for workmanship errors and/or accidents.<br/><br/>Check individual contractors to see if they also offer their own work guarantee.<br/><br/>Contact Bizza for more info. `;

        fseData.is_work_guaranteed = isJimsFse ? true : fseData.is_work_guaranteed;

        if (!fseData.is_work_guaranteed) {
            this.formElement.querySelector('.work-guarantee').classList.add('hidden');
        }

        this.formElement.querySelector('.work-guarantee-amount').innerText = isJimsFse ? '$1,000,000' : '$200';

        tooltip.innerHTML = isJimsFse ? tooltip.innerHTML : contractorTooltipHTML;
    }

    setFSERatingStars(fseData) {
        let rating = fseData.star_rating;
        this.formElement.querySelector('.rating .rating-count').innerText = rating.toFixed(2);
        this.updateRatingStars(Math.round(rating));
    }

    updateRatingStars(rating) {
        let ratingStars = this.formElement.querySelectorAll('.jims-wpcb-success-fse .rating-stars .star');
        if (ratingStars.length === 0) return;

        for (let i = 0; i < rating; i++) {
            if (i < rating) {
                ratingStars[i].classList.add('active');
            } else {
                ratingStars[i].classList.remove('active');
            }
        }
    }

    /**
     * Manages progress, steps, and navigation
     * @returns {number}
     */
    setProgress() {
        let selectedServices = this.getSelectedServices();
        let formContainer = this.formElement.querySelector('.jims-wpcb-form__wrapper');
        formContainer.scrollTop = 0;

        this.navButtons.prev.disabled = this.currentStep === 0;
        this.setNextButtonState(selectedServices);

        //set visibility of submit button
        this.navButtons.submit.classList.toggle('hidden', this.currentStep !== this.formElement.querySelectorAll(`.${this.prefixClass}__step`).length - 1);

        let steps = this.formElement.querySelectorAll(`.${this.prefixClass}__progress-step`);

        if (this.currentStep === 0) { //show first step
            this.formElement.querySelector(`.${this.prefixClass}__step`).classList.remove('hidden');
        }

        if (this.currentStep >= 0) {

            //remove the hidden in all previous steps
            for (let i = 0; i <= this.currentStep; i++) {
                steps[i].classList.remove('hidden');
            }
            // steps[this.currentStep].classList.remove('hidden');

            //add hidden to all steps after current step
            for (let i = this.currentStep + 1; i < steps.length; i++) {
                steps[i].classList.add('hidden');
            }
            // for (let i = 0; i < steps.length; i++) {
            //     if(i === this.currentStep) continue;
            //
            //     steps[i]?.classList.add('hidden');
            // }
        }

        this.googleAnalytics.sendEvent('form', 'step', 'booking-step', this.currentStep + 1);
        return this.currentStep / this.formElement.querySelectorAll(`.${this.prefixClass}__step`).length;
    }

    /**
     * Create divs inside the progress bar for each step
     */
    setupProgress() {
        let steps = this.formElement.querySelectorAll(`.${this.prefixClass}__step`);
        let progress = this.formElement.querySelector('.progress');

        //create divs for each step
        for (let i = 0; i < steps.length; i++) {
            let div = document.createElement('div');
            div.classList.add(`${this.prefixClass}__progress-step`);
            div.classList.add('hidden');
            progress.appendChild(div);
        }

        //set progress
        this.setProgress();

        if (this.options.bookingFormServiceIds && this.options.bookingFormServiceIds.length > 0) {
            this.nextStep();
            this.navButtons.prev.classList.add('hidden');

            this.formElement.addEventListener('jims-wpcb-services-loaded', () => {
                this.loadSelectedServices();
            });
        }
    }

    setBookingRequest() {
        this.bookingRequest = new BookingRequest(data.firstname, data.lastname, data.email, data.phone, data.address, data.comments, data.businessName);
    }

    setCurrentStep(step) {
        this.currentStep = step;
    }

    /**
     * Submit form
     */
    nextStep() {
        //count steps
        let steps = this.formElement.querySelectorAll(`.${this.prefixClass}__step`);

        //if current step is less than total steps
        if (this.currentStep < steps.length) {
            //hide current step
            if (steps[this.currentStep]) {
                steps[this.currentStep].classList.add('hidden');
            }

            //show next step
            steps[this.currentStep + 1].classList.remove('hidden');

            //set current step
            this.setCurrentStep(this.currentStep + 1);
            this.setProgress();
        }
    }

    /**
     * Previous step
     */
    prevStep() {
        //count steps
        let steps = this.formElement.querySelectorAll(`.${this.prefixClass}__step`);

        //if current step is greater than 1
        if (this.currentStep > 0) {
            //hide current step
            steps[this.currentStep].classList.add('hidden');

            //show previous step
            steps[this.currentStep - 1].classList.remove('hidden');

            //set current step
            this.setCurrentStep(this.currentStep - 1);
            this.setProgress();
        }
    }

    loadServices() {
        let servicesContainer = this.isServiceGroupingEnabled()
            ? this.formElement.querySelector(`.${this.prefixClass}-service-category-container`)
            : this.formElement.querySelector('.services .results ul');

        if (servicesContainer === null) {
            this.loader.removeServiceLoader();

            if(this.options.bookingFormServiceIds.length === 0) return;
        }

        let services = this.getServices();

        //load services in memory
        services.then((response) => {
            this.setFormNotification(this.messages.search);
            let hiddenServices = [];
            let visibleServices = [];

            if (this.options.grouping_enabled === '1' &&
                (this.options.bookingFormServiceIds && this.options.bookingFormServiceIds.length > 0)) {
                let serviceData = [];
                let groups = Object.values(response);
                groups.forEach((group) => {
                    let groupServices = group.services;
                    if (Array.isArray(groupServices)) {
                        //merge all groupServices into one serviceData array
                        Object.values(groupServices).forEach((service) => {
                            serviceData.push(service);
                        });
                        response= serviceData;
                    }
                });
            }

            if (this.isServiceGroupingEnabled()) {
                let serviceGroups = Object.values(response);

                //ordering groups by order property, the bigger the order the higher the group
                serviceGroups = serviceGroups.sort((a, b) => {
                    return (a.order < b.order) ? 1 : -1;
                });

                //render groups
                serviceGroups.forEach((serviceGroup) => {
                    let group = new ServiceGroup(serviceGroup.name, serviceGroup.order, serviceGroup.description, this);
                    let groupServices = serviceGroup.services;

                    if (!Array.isArray(groupServices)) {
                        groupServices = Object.values(groupServices);
                    }

                    if (this.options.alphabetical_order === '1') {
                        //sort services alphabetically
                        groupServices = this.OrderServicesAlphabetically(groupServices);
                    }

                    groupServices = this.filterServiceVisibility(groupServices);

                    if (groupServices.length > 0) {
                        groupServices.forEach((service) => {
                            visibleServices.push(service.id);

                            let newService = new Service(service.friendly_name, service.id, service?.description, true);
                            group.addService(newService);
                            this.serviceList.push(newService);
                        });

                        group.createDomElement();
                        this.groups.push(group);

                        servicesContainer.appendChild(group.element);
                    }
                });

            } else {
                let services = this.filterServiceVisibility(Object.values(response));

                services.forEach((service) => {
                    visibleServices.push(service.id);
                    this.serviceList.push(new Service(service.friendly_name, service.id));
                });

                //hidden services variable is referring to the services that are not available through the API
                //and are set in the booking form shortcode, normally they exist on FMS but are not available
                // on national level
                hiddenServices = this.options.bookingFormServiceIds.filter((id) => {
                    return visibleServices.indexOf(id) === -1;
                });

                if (hiddenServices.length > 0) {
                    hiddenServices.forEach((id) => {
                        this.serviceList.push(new Service(`Custom Service`, id));
                    });
                }

                //render services
                if (this.serviceList.length > 0) {

                    if (this.options.alphabetical_order === '1') {
                        //sort services alphabetically
                        this.serviceList = this.OrderServicesAlphabetically(this.serviceList);
                    }

                    this.serviceList.forEach((service) => {
                        if (servicesContainer !== null)
                        servicesContainer.appendChild(service.element);

                        if (this.options.bookingFormServiceIds && this.options.bookingFormServiceIds.indexOf(service.id) !== -1 && !service.selected) {
                            service.toggleSelected();
                        }
                    });

                }
            }

            this.formElement.dispatchEvent(this.servicesLoadedEvent);
        }).catch((error) => {
            log.info(error);
            this.loader.removeServiceLoader();
            this.setErrorMessageNotification({statusText: 'There was an error loading the services.'});
        });

    }

    /**
     * Filters services based on the show only services and excluded services options
     *
     * @param response
     * @return {*}
     */
    filterServiceVisibility(response) {

        //check if excluded services are set or show only services are set
        // if both the show only services and excluded services are set, the excluded services will be ignored
        if (this.options.exclude_services.length > 0 || this.options.show_only_services.length > 0) {
            let excludedServices = this.options.exclude_services;
            let showOnlyServices = this.options.show_only_services;

            //if show only services are set, remove all services that are not in the show only services array
            if (showOnlyServices.length > 0) {
                response = response.filter((service) => {
                    return showOnlyServices.indexOf(service.id) !== -1;
                });
            }

            //if excluded services are set, remove all services that are in the excluded services array
            if (excludedServices.length > 0 && showOnlyServices.length === 0) {
                response = response.filter((service) => {
                    return excludedServices.indexOf(service.id) === -1;
                });
            }
        }

        return response;
    }

    /**
     * Orders services alphabetically
     * @param serviceList
     * @return {*}
     * @constructor
     */
    OrderServicesAlphabetically(serviceList) {
        return serviceList.sort((a, b) => {
            let aName = a.name.toLowerCase();
            let bName = b.name.toLowerCase();

            if (aName < bName) return -1;
            if (aName > bName) return 1;
            return 0;
        });
    }

    /**
     * Returns true if the service grouping is enabled and no preselected services are set
     * @return {boolean}
     */
    isServiceGroupingEnabled() {
        return this.options.grouping_enabled === '1' &&
            !(this.options.bookingFormServiceIds && this.options.bookingFormServiceIds.length > 0);
    }

    loadSelectedServices() {
        let selectedServicesContainer = this.formElement.querySelector('.selected_services ul');

        this.serviceList.forEach((service) => {
            if (service.selected) {
                selectedServicesContainer.appendChild(service.selectedElement);
            } else {
                let element = selectedServicesContainer.querySelector('.service-id-' + service.id);
                if (element && element.parentNode === selectedServicesContainer) {
                    service.selectedElement.remove();
                }
            }
        });

        let selectedServices = this.getSelectedServices();
        // this.navButtons.next.disabled = this.isNextButtonDisabled(selectedServices);
        this.setNextButtonState(selectedServices);
    }

    getServices() {
        let http = new HttpRequest(this.options.rest_url);
        //create query string
        let queryString = new URLSearchParams({
            'division_id': this.options.bookingFormDivisionId,
            'use_grouping': this.options.grouping_enabled
        });

        return http.get(`${this.endpoints.services}?${queryString.toString()}`);
    }


    /**
     *
     * @param addressObject
     * @param selectedServices
     * @return {{firstname: string | number | any, address, comments: string | number | any, phone: string | number | any, businessName: string | number | any, division_id: string | number | any, services, email: string | number | any, lastname: string | number | any}}
     */
    getRequestFields(form, addressObject, selectedServices) {
        return {
            firstname: form.querySelector('input[name="firstname"]').value,
            lastname: form.querySelector('input[name="lastname"]').value,
            email: form.querySelector('input[name="email"]').value,
            phone: form.querySelector('input[name="phone"]').value,
            address: addressObject,
            comments: form.querySelector('textarea[name="comments"]').value,
            businessName: form.querySelector('input[name="businessName"]').value,
            services: selectedServices,
            division_id: form.querySelector('input[name="division"]').value
        }
    }

    /**
     *
     * @param fields
     * @param previous_response
     * @param selectedFSE
     * @return {*&{booking_id: *, previous_response: {result: boolean, payload: {avaliableFSEs: *[], confirmCodes: *, booking_request_id: *, provider_id: *}, message: string, returnType: number, statusCode: number}}}
     */
    getFinalRequestFields(fields, previous_response, selectedFSE, used_jims_before) {
        let payload = previous_response.fms_data.payload;

        return {
            ...fields,
            'booking_id': previous_response.booking_id,
            'used_jims_before': used_jims_before,
            'previous_response': {
                'result': true,
                'message': "",
                'statusCode': 0,
                'returnType': 1,
                'payload': {
                    'booking_request_id': payload.bookingRequestId,
                    'avaliableFSEs': [selectedFSE],
                    'provider_id': selectedFSE.business_id,
                    'confirmCodes': payload.confirmCodes,
                    'previousLeadId': payload.previousLeadId,
                }
            }
        }
    }

    /**
     *
     * @param endpoint
     * @returns {Promise<unknown>}
     */
    send(endpoint, data) {
        const http = new HttpRequest(this.options.rest_url);

        if (endpoint === this.endpoints.create.direct) {
            endpoint = `${endpoint}?token=bypass`;
        } else {
            endpoint = `${endpoint}&token=bypass`;
        }

        return http.post(endpoint, data);
    }

    /**
     * @returns {boolean}
     * @returns {Element | SVGSymbolElement | SVGMetadataElement | SVGUseElement | SVGAnimateElement | SVGFEImageElement | SVGPathElement | SVGViewElement | SVGFEConvolveMatrixElement | SVGFECompositeElement | SVGEllipseElement | SVGFEOffsetElement | SVGTextElement | SVGDefsElement | SVGFETurbulenceElement | SVGImageElement | SVGFEFuncGElement | SVGMPathElement | SVGTSpanElement | SVGClipPathElement | SVGLinearGradientElement | SVGFEFuncRElement | SVGScriptElement | SVGFEColorMatrixElement | SVGFEComponentTransferElement | SVGStopElement | SVGMarkerElement | SVGFEMorphologyElement | SVGFEMergeElement | SVGFEPointLightElement | SVGForeignObjectElement | SVGFEDiffuseLightingElement | SVGStyleElement | SVGFEBlendElement | SVGCircleElement | SVGPolylineElement | SVGDescElement | SVGFESpecularLightingElement | SVGLineElement | SVGFESpotLightElement | SVGFETileElement | SVGPatternElement | SVGTitleElement | SVGSwitchElement | SVGRectElement | SVGFEDisplacementMapElement | SVGFEFuncAElement | SVGFEFuncBElement | SVGFEMergeNodeElement | SVGTextPathElement | SVGFEFloodElement | SVGMaskElement | SVGAElement | SVGAnimateTransformElement | SVGSetElement | SVGSVGElement | SVGAnimateMotionElement | SVGGElement | SVGFEDistantLightElement | SVGFEDropShadowElement | SVGRadialGradientElement | SVGFilterElement | SVGPolygonElement | SVGFEGaussianBlurElement | HTMLSelectElement | HTMLLegendElement | HTMLElement | HTMLTableCaptionElement | HTMLTextAreaElement | HTMLModElement | HTMLHRElement | HTMLOutputElement | HTMLEmbedElement | HTMLCanvasElement | HTMLScriptElement | HTMLInputElement | HTMLMetaElement | HTMLStyleElement | HTMLObjectElement | HTMLTemplateElement | HTMLBRElement | HTMLAudioElement | HTMLIFrameElement | HTMLMapElement | HTMLTableElement | HTMLAnchorElement | HTMLMenuElement | HTMLPictureElement | HTMLParagraphElement | HTMLTableCellElement | HTMLTableSectionElement | HTMLQuoteElement | HTMLProgressElement | HTMLLIElement | HTMLTableRowElement | HTMLSpanElement | HTMLTableColElement | HTMLOptGroupElement | HTMLDataElement | HTMLDListElement | HTMLFieldSetElement | HTMLSourceElement | HTMLBodyElement | HTMLDivElement | HTMLUListElement | HTMLDetailsElement | HTMLHtmlElement | HTMLAreaElement | HTMLPreElement | HTMLMeterElement | HTMLOptionElement | HTMLImageElement | HTMLLinkElement | HTMLHeadingElement | HTMLSlotElement | HTMLVideoElement | HTMLTitleElement | HTMLButtonElement | HTMLHeadElement | HTMLDialogElement | HTMLTrackElement | HTMLOListElement | HTMLDataListElement | HTMLLabelElement | HTMLFormElement | HTMLTimeElement | HTMLBaseElement}
     */
    hasServices() {
        return this.formElement.querySelector('[name="services"]:checked');
    }
}

export default BookingForm;