<?php

namespace JimsCustomerBooking\RestfulAPI;

use JimsCustomerBooking\Admin\BackendPages\NativeFormSettingsPage;
use JimsCustomerBooking\BookingForm;
use JimsCustomerBooking\Database\Booking;
use JimsCustomerBooking\Generators\CBURLGenerator;
use JimsCustomerBooking\Helpers\Http;
use JimsCustomerBooking\Helpers\OptionHelper;
use JimsCustomerBooking\Helpers\StringHelper;
use JimsCustomerBooking\Integrations\ANotification;
use JimsCustomerBooking\Integrations\JimsFmsBooking;
use JimsCustomerBooking\Integrations\SlackNotification;
use JimsGroup\Apis\CustomerBooking;
use JimsGroup\Datatypes\BookingResponseType;
use WpOrg\Requests\Exception;
use function categoryPosts\register_route;

class CustomerBookingController extends RestController implements IRestfulController
{
    protected $route = 'customer-booking';

    protected $findajim_autocomplete_token = '84bcb06b2d534989cf811f250b26dc10';

    /**
     * @var string $findajim_api_mode
     */
    protected $findajim_api_mode = 'production'; // 'production', 'development

    /**
     * @var string[] $findajim_endpoints_base_url
     */
    protected $findajim_endpoints_base_url = [
        'development' => 'http://127.0.0.1:8001',
        'production' => 'https://findajim.jims.net',
    ];

    /**
     * @var string[] $findajim_endpoints
     */
    protected $findajim_endpoints = [
        'autocomplete' => '/api/address/autocomplete',
        'autocomplete_details' => '/api/address/autocomplete?with_details=1',
        'details' => '/api/address/details',
    ];

    protected $messages = [
        'success' => 'Success! Your booking is confirmed.',
        'error' => 'Booking unsuccessful. Please try again.'
    ];


    private $available_modes = [
        'production', 'uat', 'hf', 'hf2', 'development'
    ];

    private $mode = 'production'; // 'production', 'uat', 'hf', 'hf2', 'development

    public function register_routes()
    {
        register_rest_route ($this->get_namespace (), $this->get_route () . '/cb-link', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => array($this, 'get_cb_link'),
                'permission_callback' => '__return_true',
            ]
        ]);

        register_rest_route ($this->get_namespace (), $this->get_route () . '/list', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => array($this, 'get_bookings'),
                'permission_callback' => [$this, 'checkPermissions'],
            ]
        ]);

        register_rest_route ($this->get_namespace (), $this->get_route () . '/create', [
            [
                'methods' => \WP_REST_Server::CREATABLE,
                'callback' => array($this, 'create_booking'),
                'permission_callback' => [$this, 'checkPermissions'],
            ]
        ]);

        register_rest_route($this->get_namespace(), $this->get_route() . '/url/fse-photo', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => array($this, 'get_fse_photo'),
                'permission_callback' => '__return_true',
            ]
        ]);

        register_rest_route($this->get_namespace(), $this->get_route() . '/validate', [
            [
                'methods' => \WP_REST_Server::CREATABLE,
                'callback' => array($this, 'validate_booking'),
                'permission_callback' => [$this, 'checkPermissions'],
            ]
        ]);

        register_rest_route ($this->get_namespace (), $this->get_route () . '/single', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => array($this, 'get_booking'),
                'permission_callback' => '__return_true',
            ]
        ]);

        register_rest_route ($this->get_namespace (), $this->get_route () . '/address', [
            [
                'methods' => \WP_REST_Server::CREATABLE,
                'callback' => array($this, 'get_address')
            ]
        ]);

        register_rest_route ($this->get_namespace (), $this->get_route () . '/address/details', [
            [
                'methods' => \WP_REST_Server::CREATABLE,
                'callback' => array($this, 'get_address_details'),
            ]
        ]);

    }

    /**
     * @param $request \WP_REST_Request
     * @return \WP_REST_Response
     */
    public function get_fse_photo(\WP_REST_Request $request){
        $fse_business_id = $request->get_param ('fse_business_id');
        $fmsBooking = new JimsFmsBooking('AU', $this->getMode ());
        $api_url = $fmsBooking->getApiUrl () . '/api/CustomerBookings/get-fse-photo/' . $fse_business_id;

        return $this->response([
            'status' => 'success',
            'url' => $api_url
        ]);
    }

    public function get_cb_link($request)
    {
        $properties = \JimsCustomerBooking\Jims_Plugin::getOptions ()->getProperties ();

        if (count ($properties) == 0) {
            return $this->response ([
                'status' => 'error',
                'message' => 'Please set up the plugin first'
            ]);
        }

        $cb_link = new CBURLGenerator($properties);
        return $this->response ([
            'status' => 'success',
            'cb_link' => $cb_link->generate (false)
        ]);
    }

    /**
     * @param $request \WP_REST_Request
     * @return \WP_REST_Response
     */
    public function create_booking($request)
    {
        $data = json_decode ($request->get_body (), true);

        if ($request->has_param ('booking_step') && $request->get_param ('booking_step') == 1) {
            return $this->do_create_booking_allocation_step ($data);
        }

        if ($request->has_param ('booking_step') && $request->get_param ('booking_step') == 2) {
            return $this->do_create_booking_final_step ($data);
        }

        return $this->create_direct_booking ($data);
    }

    /**
     * @param $request \WP_REST_Request
     * @return \WP_REST_Response
     */
    public function validate_booking(\WP_REST_Request $request){
        $this->validate(json_decode ($request->get_body (), true), $errors);
        return $this->response ([
            'status' => count($errors) == 0 ? 'success' : 'error',
            'errors' => $errors
        ]);
    }


    /**
     * @param $data array
     * @return \WP_REST_Response
     */
    protected function do_create_booking_allocation_step($data)
    {
        $result = false;
        $fmsData = [];
        $notification_sent = false;
        $errors = [];

        if (!$this->validate ($data, $errors)) {
            return $this->response (
                $this->getResponseArray ('error', false, [ 'error' => $errors], false, $this->messages)
            );
        }

        list($booking, $booking_id, $data) = $this->saveLead ($data);

        //Finalize booking to FMS
        try {
            $FMSBooking = new JimsFmsBooking(
                $data['address']['country'],
                $this->getMode (),
            );

            //replace suburb for new zealand
            if($data['address']['country'] === 'NZ'){
                $parts = explode(',', $data['address']["full_address"]);
                if (isset($parts[1])) {
                    $data['address']["suburb"] = trim($parts[1]);
                }

                //replace Road in address street_name with Rd
                $data['address']["street_name"] = str_replace('Road', 'Rd', $data['address']["street_name"]);

//                $data['address']["state"] = 'NZ';
//                $data['address']['postcode'] = 'NZ ' . $data['address']['postcode'];
            }

            error_log(json_encode($data), 0);

            extract( $this->formatBookingFormData ($data) );

            $booking_form  = new BookingForm(
                $customer,
                $address,
                $division_id,
                $services
            );

            $response = $FMSBooking->doAllocationRequest($booking_form);
            $fmsData = $response->toArray ();

            //check if booking was direct
            if($response->isBookingCompleted ()){
                $result = $this->validateBookingResponse($response);
                if($result) $this->updateBookingStatus($booking, $booking_id, true);
            }else {
                //check if fmsData has error code
                if($this->responseHasError($response)) {
                    $this->updateBookingStatus($booking, $booking_id, false, $response->getPayloadProperty('bookingConfirmCode'));
                }

                //if not direct, respond to the frontend with the data
                return  $this->response([
                    'status' => 'step',
                    'result' => true,
                    'fms_data' => $fmsData,
                    'booking_id' => $booking_id,
                ]);
            }

        } catch (Exception $e) {
            //in case of error, respond to the frontend with the data
            $this->messages['error'] = $e->getMessage ();
            return $this->response (
                $this->getResponseArray ('error', false, $fmsData, false, $this->messages)
            );
        }

        // Send notification
        $notification_sent = $this->sendSlackNotification ($booking_id, $result);

        //if the booking is completed, update booking status and respond
        return $this->response (
            $this->getResponseArray ($result ? 'success' : 'error', $result, $fmsData, $notification_sent, $this->messages)
        );
    }

    /**
     * @param $data array
     * @return \WP_REST_Response
     */
    protected function do_create_booking_final_step($data)
    {
        $result = false;
        $notification_sent = false;
        $fmsData = [];
        $errors = [];

        $this->validate ($data, $errors);
        if(!isset($data['previous_response'])){
            $errors['previous_response'] = true;
        }

        $valid = empty($errors);

        if (!$valid) {
            return $this->response (
                $this->getResponseArray ('error', false, ['errors' => $errors], false, $this->messages)
            );
        }
        error_log(json_encode($data), 0);

//        $booking = $booking_id = null;
        extract( $this->formatBookingFormData ($data) );

        //Finalize booking to FMS
        try {
            $FMSBooking = new JimsFmsBooking(
                $data['address']['country'],
                $this->getMode (),
            );
            $used_jims_before = isset($data['used_jims_before']) ? $data['used_jims_before'] : false;

            $booking_form  = new BookingForm(
                $customer,
                $address,
                $division_id,
                $services
            );

            $PreviousResponse = new BookingResponseType( json_encode($data['previous_response']) );

            error_log( 'Used Jims Before: ' . $used_jims_before, 0);
            $response = $FMSBooking->doFinalRequest ($booking_form, $PreviousResponse, $used_jims_before);
            $fmsData = $response->toArray ();

            if($response->isBookingCompleted ()){
                $result = $this->validateBookingResponse($response);
            }

            //check if fmsData has error code
            if($this->responseHasError($response)) {
                $this->updateBookingStatus($booking, $booking_id, false, $response->getPayloadProperty('bookingConfirmCode'));
            }

        } catch (Exception $e) {
            $this->messages['error'] = $e->getMessage ();

            //check if fmsData has error code
            if($this->responseHasError($response)) {
                $this->updateBookingStatus($booking, $booking_id, false, $response->getPayloadProperty('bookingConfirmCode'));
            }

            return $this->response (
                $this->getResponseArray ('error', false, $fmsData, false, $this->messages)
            );
        }


        // Update booking status
        if (isset($data['booking_id']) && !empty($data['booking_id'])) {
            $booking = new Booking();
            $this->updateBookingStatus ($booking, $data['booking_id'], $result);

            // Send notification
            $notification_sent = $this->sendSlackNotification ($data['booking_id'], $result);
        }

        return $this->response (
            $this->getResponseArray (
                $result ? 'success' : 'error',
                $result, $fmsData, $notification_sent, $this->messages)
        );
    }

    /**
     * @param array $data
     * @return \WP_REST_Response
     */
    protected function create_direct_booking(array $data)
    {
        $validation_errors = [];
        if (!$this->validate ($data, $validation_errors)) {
            return $this->response (
                $this->getResponseArray ('error', false, [ 'errors' => $validation_errors ], false, $this->messages)
            );
        }

        list($booking, $booking_id, $data) = $this->saveLead ($data);

        // Send to FMS
        list($FMSBooking, $result) = $this->connectToFMSAndSendData ($data, $booking, $booking_id);

        $fmsData = $FMSBooking->getMessages (); //GET FMS RESPONSES
        list($result) = $this->updateMessages($fmsData, $result);

        // Send notification
        $notification_sent = $this->sendSlackNotification ($booking_id, $result);

        //respond to the frontend with the data
        return $this->response (
            $this->getResponseArray ($result ? 'success' : 'error', $result, $fmsData, $notification_sent, $this->messages)
        );
    }

    /**
     * @param $lastMessage
     * @return string
     */
    public function getLastMessage(array $fmsData){
        //get last message
        $lastMessageData = $this->getTheLastNotEmptyItem($fmsData, true);

        $lastMessage = StringHelper::isJson(  $lastMessageData ) ? json_decode($lastMessageData, true) : $lastMessageData;

        error_log (__FILE__ . ' ' . var_export($lastMessage, true), 0); //log last response as final response

        if( is_array($lastMessage) ){
            return $lastMessage['message'] ?? '';
        }

        return $lastMessage;
    }

    /**
     * @param $status
     * @param $result
     * @param $lastMessage
     * @param $notification_sent
     * @param $messages
     * @return array
     */
    protected function getResponseArray($status, $result, $lastMessage, $notification_sent, $messages)
    {
        if (!is_bool ($result)) {
            $result = false;
        }

        return [
            'status' => $status,
            'result' => $result,
            'fms_data' => $lastMessage,
            'notification' => $notification_sent ? 'sent' : 'not sent',
            'notification_message' => $result ? $messages['success'] : $messages['error'],
        ];
    }

    /**
     * @param $array
     * @return array|null
     */
    protected function getTheLastNotEmptyItem($array, $onlyArrays = true)
    {
        $last = null;
        foreach ($array as $item) {

            $array_result = true;
            if($onlyArrays){
                $array_result = is_array($item);
            }

            if ( $array_result && !empty($item)) {
                $last = $item;
            }
        }

        return $last;
    }

    /**
     * @param $request
     * @return \WP_REST_Response
     */
    public function get_bookings($request)
    {
        $page = $limit = $status = null;
        $values = $this->getSearchParams ($request);

        extract ($values);

        //create transient key
        $transient_key = StringHelper::GetPrefixedFromParts (['customer_booking_list', $page, $limit, $status], 'machine');


        if (false === ($result = get_transient ($transient_key))) {
            $booking = new Booking();
            $data = $booking->getPage ($page, $limit, $status);

            $result = array_map (function ($item) {
                return [
                    'id' => $item['booking_id'],
                    'data' => json_decode ($item['data']),
                    'status' => $item['booking_status'],
                    'created_at' => $item['created_at'],
                    'updated_at' => $item['updated_at']
                ];
            }, $data);

            set_transient ($transient_key, $result, 0.5 * HOUR_IN_SECONDS);
        }

        return $this->response ([
            'status' => 'success',
            'page_count' => Booking::getPageCount ($limit, $status),
            'current_page' => $page,
            'data' => $result
        ]);
    }

    /**
     * @param $request \WP_REST_Request
     * @return \WP_REST_Response
     */
    public function get_booking($request)
    {
        $booking = new Booking();

        $id = $request->get_param ('booking_id');
        $data = $booking->getFromId ($id);

        return $this->response ($data);
    }

    /**
     * @param $request
     * @return array|int[]
     */
    protected function getSearchParams($request): array
    {
        $params_defaults = [
            'page' => 1,
            'limit' => 10,
            'status' => 'all'
        ];

        $data = $request->get_params ();
        $search_params = \array_keys ($params_defaults);

        $values = \array_combine ($search_params, \array_map (function ($key) use ($data, $params_defaults) {
            return isset($data[$key]) ? $data[$key] : $params_defaults[$key];
        }, $search_params));

        return $values;
    }

    /**
     * @param $request \WP_REST_Request
     * @return mixed
     */
    public function get_address($request)
    {
        $address = $request->get_param ('address');
        $country_code = $request->get_param ('country_code');

        $key = StringHelper::GetPrefixedFromParts (['booking', 'address', $address], 'machine');

        $response = $this->getResponse ($key, [
            'address' => $address,
            'country_code' => $country_code
        ], 0.5 * HOUR_IN_SECONDS, 'get_address_suggestions');

        return json_decode ($response['result']);
    }

    public function get_address_suggestions($address, $country_code = null)
    {
        return Http::post ($this->getFindAJimEndpoint ('autocomplete'), [
            'address' => $address,
            'countryCode' => empty($country_code) ? 'AU' : $country_code,
            'token' => $this->findajim_autocomplete_token
        ]);
    }

    /**
     * @param $request
     * @return mixed
     */
    public function get_address_details($request)
    {
        $address = $request->get_param ('address');

        $key = StringHelper::GetPrefixedFromParts (['booking', 'address', 'details', $address], 'machine');

        if (false === ($response = get_transient ($key))) {

            $response = Http::post ($this->getFindAJimEndpoint ('details'), [
                'address' => $address,
                'token' => $this->findajim_autocomplete_token
            ]);

            set_transient ($key, $response, 0.5 * HOUR_IN_SECONDS);
        }

        return json_decode ($response['result']);
    }

    /**
     * @param mixed $data
     * @param Booking $booking
     * @param int $booking_id
     * @return array
     */
    protected function connectToFMSAndSendData(mixed $data, Booking $booking, int $booking_id): array
    {
        $result = false;
        //Connect to FMS and send data
        try {
            $FMSBooking = new JimsFmsBooking(
                $data['address']['country'],
                $this->getMode (),
            );

            error_log (__FILE__ . ' ' . json_encode ($data), 0);

            extract( $this->formatBookingFormData ($data) );

            $result = $FMSBooking->doBookingRequest (
                $customer,
                $address,
                $division_id,
                $services
            );
        } catch (\Exception $e) {
            error_log(__FILE__ . ' ' . $e->getMessage(), 0);
            $result = false;
        }

        $this->updateBookingStatus ($booking, $booking_id, $result);

        return array($FMSBooking, $result);
    }

    /**
     * @return string
     */
    public function getMode()
    {   $settings = OptionHelper::get(NativeFormSettingsPage::$option_name, [] );

        if(isset($settings['fms_environment']) && in_array ($settings['fms_environment'], $this->available_modes)) {
            return $settings['fms_environment'];
        }

        return $this->mode;
    }

    /**
     * @param $fmsData
     * @param bool $result
     * @return array
     */
    public function updateMessages($fmsData, bool $result): array
    {
        if (count($fmsData) > 0) {

            $fmsData = array_map(function ($item) {
                return StringHelper::isJson($item) ? json_decode($item, true) : $item;
            }, $fmsData);

            $lastMessage = $this->getLastMessage($fmsData);

            if (strtolower($lastMessage) !== 'booked') {
                $result = false;
            }

            $this->messages['error'] = $lastMessage ?? $this->messages['error'];
        }
        return array($result);
    }

    /**
     * @param BookingResponseType $response
     * @param mixed $booking
     * @param mixed $booking_id
     * @return true
     * @throws Exception
     */
    public function validateBookingResponse(BookingResponseType $response): bool
    {
        if ($response->getPayloadProperty('bookingConfirmCode') === BookingResponseType::CONFIRM_CODE_BOOKED) {
            error_log('Booking completed', 0);
        } else {
            error_log($response->getStatusMessage());
            throw new Exception($response->getStatusMessage(), 0);
        }

        return true;
    }

    /**
     * @param $response
     * @return bool
     */
    public function responseHasError($response){
        $responseCode = (int) $response->getPayloadProperty('bookingConfirmCode');
        return $responseCode > 0 && $responseCode <= 9;
    }

    /**
     * @param $data
     * @return BookingForm
     */
    protected function getBookingForm($data){
        list($customer, $address, $division_id, $service_ids) = $this->formatBookingFormData ($data);
        return new BookingForm($customer, $address, $division_id, $service_ids);
    }

    /**
     * @param $data
     * @return array
     */
    protected function formatBookingFormData($data)
    {
        return [
            'customer' => [
                'address' => null,
                'addressString' => $data['address']['full_address'],
                'allow_rating' => true,
                'comments' => $data['comments'],
                'isUserInfoConfirmed' => null,
                'emailAddress' => $data['email'],
                'firstName' => $data['firstname'],
                'lastName' => $data['lastname'],
                'phoneNumber' => $data['phone'],
                'subscribed' => true,
                'businessName' => null
            ],
            'address' => $data['address'],
            'division_id' => $data['division_id'],
            'services' => $data['services'],
        ];
    }

    /**
     * @param int $booking_id
     * @param mixed $result
     * @return bool
     */
    protected function sendSlackNotification(int $booking_id, mixed $result): bool
    {
        $notification = new SlackNotification('https://hooks.slack.com/services/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', [
            'error_id' => sprintf ('id %s', "{$booking_id}"),
            'booking_id' => sprintf ('id %s', "{$booking_id}"),
        ]);

        $notification_sent = $notification->send ($result ? ANotification::NOTIFICATION_TYPE_INFO : ANotification::NOTIFICATION_TYPE_ERROR);
        return $notification_sent;
    }

    /**
     * @param $endpoint_name
     * @return string
     */
    protected function getFindAJimEndpoint($endpoint_name): string
    {
        return sprintf ('%s%s',
            $this->findajim_endpoints_base_url[$this->findajim_api_mode],
            $this->findajim_endpoints[$endpoint_name]
        );
    }

    /**
     * @param string $key
     * @param array $params
     * @param int $cache_ttl
     * @param string $callback_name
     * @return array|mixed
     */
    protected function getResponse(string $cache_key, array $params, $cache_ttl, $callback_name): mixed
    {

        $cached_response = $response = get_transient ($cache_key);
        if (false === ($cached_response)) {
//            $response = $this->get_address_suggestions ($address, $country_code);
            $response = call_user_func_array ([$this, $callback_name], $params);
            set_transient ($cache_key, $response, $cache_ttl);
        }

        if ($this->stored_response_is_empty ($cached_response)) {
            $response = call_user_func_array ([$this, $callback_name], $params);
        }
        return $response;
    }

    protected function validate(array $data, &$errors = [])
    {
        $required_fields = [
            'firstname',
            'lastname',
            'email',
            'phone',
            'address',
            'division_id',
            'services',
            'comments'
        ];

        foreach ($required_fields as $field) {
            if($field === 'comments' && isset($data[$field])){
                continue;
            }

            if (empty($data[$field])) {
                $errors[$field] = true;
            }
        }

        if (empty($errors) && !$this->validateAddressObject($data['address'])) {
            $errors['address'] = true;
        }

        return empty($errors);
    }

    protected function validateAddressObject(array $address)
    {
        $required_fields = [
            'full_address',
            'country',
            'state',
            'suburb',
            'postcode',
            'street_number',
            'street_name',
            'geocode_y',
            'geocode_x',
        ];

        $is_valid = true;
        foreach ($required_fields as $field) {
            if (empty($address[$field])) {
                $is_valid = false;
                break;
            }
        }
        return $is_valid;
    }

    /**
     * @param array $data
     * @return array
     */
    protected function saveLead(array $data): array
    {
        $booking = new Booking();
        $booking_id = $booking->add ($data);

        if (isset($data['address'])) {
            unset($data['address']['place_id']);
            unset($data['address']['unit_number']);
        }
        return array($booking, $booking_id, $data);
    }

    /**
     * @param Booking $booking
     * @param int $booking_id
     * @param bool $result
     * @return void
     */
    protected function updateBookingStatus(Booking $booking, int $booking_id, bool $result = false, $error_code = null): void
    {
        $booking->updateStatus ($booking_id, $result ? 'sended' : 'pending', $error_code);
    }
}